from typing import List, Optional

from starlette.requests import Request

from watchmen_auth import AuthenticationManager, AuthFailOn401, AuthFailOn403, PrincipalService
from watchmen_model.admin import UserRole
from .authentication import get_admin_principal_by, get_any_admin_principal_by, get_any_principal_by, \
	get_console_principal_by, get_principal_by_jwt as by_jwt, get_principal_by_pat as by_pat, \
	get_super_admin_principal_by
from .util import raise_401, raise_403


# singleton
class AuthenticationManagerWrapper:
	authenticationManager: Optional[AuthenticationManager] = None


WRAPPER = AuthenticationManagerWrapper()


def register_authentication_manager(authentication_manager: AuthenticationManager) -> None:
	WRAPPER.authenticationManager = authentication_manager


def retrieve_authentication_manager() -> Optional[AuthenticationManager]:
	return WRAPPER.authenticationManager


def get_principal_by_jwt(token: str, roles: List[UserRole]) -> PrincipalService:
	try:
		return by_jwt(WRAPPER.authenticationManager, token, roles)
	except AuthFailOn401:
		raise_401('Unauthorized visit.')
	except AuthFailOn403:
		raise_403()


def get_principal_by_pat(token: str, roles: List[UserRole]) -> PrincipalService:
	try:
		return by_pat(WRAPPER.authenticationManager, token, roles)
	except AuthFailOn401:
		raise_401('Unauthorized visit.')
	except AuthFailOn403:
		raise_403()


def get_super_admin_principal(request: Request) -> PrincipalService:
	return get_super_admin_principal_by(WRAPPER.authenticationManager)(request)


def get_any_admin_principal(request: Request) -> PrincipalService:
	return get_any_admin_principal_by(WRAPPER.authenticationManager)(request)


def get_admin_principal(request: Request) -> PrincipalService:
	return get_admin_principal_by(WRAPPER.authenticationManager)(request)


def get_console_principal(request: Request) -> PrincipalService:
	return get_console_principal_by(WRAPPER.authenticationManager)(request)


def get_any_principal(request: Request) -> PrincipalService:
	return get_any_principal_by(WRAPPER.authenticationManager)(request)



from __future__ import annotations

from abc import abstractmethod
from logging import getLogger
from typing import Callable, List, Optional, Tuple

from fastapi import FastAPI

from watchmen_auth import AuthenticationProvider
from watchmen_model.admin import User
from .auth_helper import register_authentication_manager
from .authentication import build_authentication_manager
from .cors import install_cors
from .prometheus import install_prometheus
from .settings import RestSettings

logger = getLogger(f'app.{__name__}')


class RestApp:
	def __init__(self, settings: RestSettings):
		self.settings = settings
		logger.info(f'Application settings[{settings.dict()}].')

	def get_settings(self) -> RestSettings:
		return self.settings

	def construct(self) -> FastAPI:
		app = FastAPI(
			title=self.settings.APP_NAME,
			version=self.settings.VERSION,
			description=self.settings.DESCRIPTION
		)

		self.init_cors(app)
		self.init_prometheus(app)

		self.init_authentication()
		self.post_construct(app)

		logger.info('REST app constructed.')
		return app

	def is_cors_on(self) -> bool:
		return self.settings.CORS

	def init_cors(self, app: FastAPI) -> None:
		if self.is_cors_on():
			install_cors(app, self.settings)

	def is_prometheus_on(self) -> bool:
		return self.settings.PROMETHEUS

	def init_prometheus(self, app: FastAPI) -> None:
		if self.is_prometheus_on():
			install_prometheus(app, self.settings)

	# noinspection PyMethodMayBeStatic
	def get_authentication_providers(self) -> List[AuthenticationProvider]:
		"""
		default return empty list, override me and return your authentication providers here.
		"""
		return []

	def init_authentication(self) -> None:
		register_authentication_manager(build_authentication_manager(
			self.settings, self.build_find_user_by_name(), self.build_find_user_by_pat(),
			self.get_authentication_providers()
		))

	def get_jwt_params(self) -> Tuple[str, str]:
		"""
		Tuple(secret_key, algorithm)
		"""
		return self.settings.JWT_SECRET_KEY, self.settings.JWT_ALGORITHM

	def get_access_token_expires_in(self) -> int:
		"""
		in minutes
		"""
		return self.settings.ACCESS_TOKEN_EXPIRE_MINUTES

	@abstractmethod
	def build_find_user_by_name(self) -> Callable[[str], Optional[User]]:
		pass

	@abstractmethod
	def build_find_user_by_pat(self) -> Callable[[str], Optional[User]]:
		pass

	@abstractmethod
	def post_construct(self, app: FastAPI) -> None:
		pass

	@abstractmethod
	def on_startup(self, app: FastAPI) -> None:
		pass



from fastapi import FastAPI

from .settings import RestSettings


def install_prometheus(app: FastAPI, settings: RestSettings) -> None:
	from starlette_prometheus import metrics, PrometheusMiddleware
	app.add_middleware(PrometheusMiddleware)
	app.add_route(settings.PROMETHEUS_CONTEXT, metrics)



from typing import Set

from pydantic import BaseSettings

DEV = "dev"
PROD = "production"


class RestSettings(BaseSettings):
	"""
	REST settings will not construct by itself, it should be inherited and constructed by inheriting one
	"""
	APP_NAME: str = 'Watchmen REST App',
	VERSION: str = '16.0.0'
	DESCRIPTION: str = 'A lighter platform for data analytics'

	API_VERSION_STR: str = ''

	# noinspection SpellCheckingInspection
	JWT_SECRET_KEY: str = '801GtEAdlE8o-iZRLBMgz30PGE_zxry82EaUYMAhNq8'
	JWT_ALGORITHM: str = 'HS256'
	ACCESS_TOKEN_EXPIRE_MINUTES: int = 60 * 24 * 8

	CORS: bool = True
	CORS_ALLOWED_ORIGINS: Set[str] = ['*']
	CORS_ALLOW_CREDENTIALS: bool = True
	CORS_ALLOWED_METHODS: Set[str] = ['*']
	CORS_ALLOWED_HEADERS: Set[str] = ['*']

	PROMETHEUS: bool = False
	PROMETHEUS_CONTEXT: str = '/metrics'

	class Config:
		# secrets_dir = '/var/run'
		env_file = '.env'
		env_file_encoding = 'utf-8'
		case_sensitive = True



from typing import Union

from watchmen_auth import PrincipalService
from watchmen_model.common import TenantBasedTuple, UserBasedTuple
from watchmen_rest.util.raise_http_exception import raise_400, raise_403
from watchmen_utilities import is_blank, is_not_blank


def validate_tenant_id(a_tuple: Union[TenantBasedTuple, UserBasedTuple], principal_service: PrincipalService) -> None:
	"""
	validate tenant id of tuple.\n
	a. for super admin, tenant id is required,\
	b. for not super admin, tenant id must be same as current principal,\n
	c. for not super admin and no tenant id, set as current principal.
	"""
	tenant_id = a_tuple.tenantId
	if principal_service.is_super_admin():
		if is_blank(tenant_id):
			raise_400('Tenant id is required.')
	elif is_not_blank(tenant_id):
		if tenant_id != principal_service.get_tenant_id():
			raise_403()
	else:
		# assign tenant id by current principal
		a_tuple.tenantId = principal_service.get_tenant_id()



from datetime import datetime, timedelta
from logging import getLogger
from typing import Callable, List, Optional, Tuple

from jose import JWTError
from jose.jwt import decode, encode
from jsonschema.exceptions import ValidationError
from starlette.requests import Request

from watchmen_auth import AuthenticationDetails, AuthenticationManager, AuthenticationProvider, AuthenticationScheme, \
	AuthFailOn401, AuthFailOn403, authorize, authorize_token, PrincipalService
from watchmen_model.admin import User, UserRole
from watchmen_utilities import ArrayHelper, is_not_blank
from .settings import RestSettings
from .util import raise_401, raise_403

logger = getLogger(__name__)


def create_jwt_token(subject: str, expires_delta: timedelta, secret_key: str, algorithm: str) -> str:
	to_encode = {'exp': datetime.now() + expires_delta, 'sub': subject}
	encoded_jwt = encode(to_encode, secret_key, algorithm=algorithm)
	return encoded_jwt


def validate_jwt(token, secret_key: str, algorithm: str):
	try:
		return decode(token, secret_key, algorithms=[algorithm])
	except (JWTError, ValidationError):
		raise AuthFailOn401('Unauthorized caused by unrecognized token.')


class JWTAuthenticationProvider(AuthenticationProvider):
	def __init__(self, secret_key: str, algorithm: str, find_user_by_name: Callable[[str], Optional[User]]):
		self.secretKey = secret_key
		self.algorithm = algorithm
		self.find_user_by_name = find_user_by_name

	def accept(self, details: AuthenticationDetails) -> bool:
		return is_not_blank(details.scheme) and details.scheme.lower() == AuthenticationScheme.JWT.value.lower()

	def authenticate(self, details: AuthenticationDetails) -> Optional[User]:
		try:
			token = details.token
			payload = validate_jwt(token, self.secretKey, self.algorithm)
			username = payload['sub']
			return self.find_user_by_name(username)
		except AuthFailOn401 as e:
			logger.error(e, exc_info=True, stack_info=True)
			return None
		except Exception as e:
			logger.error(e, exc_info=True, stack_info=True)
			raise AuthFailOn401('Unauthorized visit.')


class PATAuthenticationProvider(AuthenticationProvider):
	def __init__(self, find_user_by_pat: Callable[[str], Optional[User]]):
		self.find_user_by_pat = find_user_by_pat

	def accept(self, details: AuthenticationDetails) -> bool:
		return is_not_blank(details.scheme) and details.scheme.lower() == AuthenticationScheme.PAT.value.lower()

	def authenticate(self, details: AuthenticationDetails) -> Optional[User]:
		token = details.token
		try:
			return self.find_user_by_pat(token)
		except AuthFailOn401 as e:
			logger.error(e, exc_info=True, stack_info=True)
			return None
		except Exception as e:
			logger.error(e, exc_info=True, stack_info=True)
			raise AuthFailOn401('Unauthorized visit.')


def build_authentication_manager(
		settings: RestSettings,
		find_user_by_name: Callable[[str], Optional[User]],
		find_user_by_pat: Callable[[str], Optional[User]],
		authentication_providers: List[AuthenticationProvider]
) -> AuthenticationManager:
	authentication_manager = AuthenticationManager()
	authentication_manager.register_provider(
		JWTAuthenticationProvider(settings.JWT_SECRET_KEY, settings.JWT_ALGORITHM, find_user_by_name)
	).register_provider(PATAuthenticationProvider(find_user_by_pat))

	ArrayHelper(authentication_providers) \
		.reduce(lambda x: authentication_manager.register_provider(x), authentication_manager)

	return authentication_manager


def parse_token(request: Request) -> Tuple[str, str]:
	authorization: str = request.headers.get("Authorization")

	if not authorization:
		raise AuthFailOn401('Unauthorized caused by token not found.')
	else:
		scheme, _, param = authorization.partition(" ")
		token = param
		return scheme, token


def get_principal_by_jwt(
		authentication_manager: AuthenticationManager, token: str, roles: List[UserRole]) -> PrincipalService:
	return authorize_token(authentication_manager, AuthenticationScheme.JWT.value, token, roles)


def get_principal_by_pat(
		authentication_manager: AuthenticationManager, token: str, roles: List[UserRole]) -> PrincipalService:
	return authorize_token(authentication_manager, AuthenticationScheme.PAT.value, token, roles)


def get_principal_by(
		authentication_manager: AuthenticationManager, roles: List[UserRole]
) -> Callable[[Request], PrincipalService]:
	def get_principal(request: Request) -> PrincipalService:
		try:
			scheme, token = parse_token(request)
			return authorize(authentication_manager, roles)(scheme, token)
		except AuthFailOn401:
			raise_401('Unauthorized visit.')
		except AuthFailOn403:
			raise_403()

	return get_principal


def get_any_principal_by(authentication_manager: AuthenticationManager) -> Callable[[Request], PrincipalService]:
	"""
	console + admin + super admin
	"""
	return get_principal_by(authentication_manager, [UserRole.SUPER_ADMIN, UserRole.ADMIN, UserRole.CONSOLE])


def get_console_principal_by(authentication_manager: AuthenticationManager) -> Callable[[Request], PrincipalService]:
	"""
	console + admin
	"""
	return get_principal_by(authentication_manager, [UserRole.ADMIN, UserRole.CONSOLE])


def get_admin_principal_by(authentication_manager: AuthenticationManager) -> Callable[[Request], PrincipalService]:
	"""
	admin only
	"""
	return get_principal_by(authentication_manager, [UserRole.ADMIN])


def get_any_admin_principal_by(authentication_manager: AuthenticationManager) -> Callable[[Request], PrincipalService]:
	"""
	super admin + admin
	"""
	return get_principal_by(authentication_manager, [UserRole.SUPER_ADMIN, UserRole.ADMIN])


def get_super_admin_principal_by(authentication_manager: AuthenticationManager) -> Callable[
	[Request], PrincipalService]:
	"""
	super admin only
	"""
	return get_principal_by(authentication_manager, [UserRole.SUPER_ADMIN])



from fastapi import FastAPI

from .settings import RestSettings


def install_cors(app: FastAPI, settings: RestSettings) -> None:
	from fastapi.middleware.cors import CORSMiddleware

	app.add_middleware(
		CORSMiddleware,
		allow_origins=settings.CORS_ALLOWED_ORIGINS,
		allow_credentials=settings.CORS_ALLOW_CREDENTIALS,
		allow_methods=settings.CORS_ALLOWED_METHODS,
		allow_headers=settings.CORS_ALLOWED_HEADERS,
	)



from logging import getLogger
from typing import Any, Optional

from fastapi import HTTPException
from starlette import status

from watchmen_utilities import is_not_blank

logger = getLogger(__name__)


def log_detail(detail: Optional[Any]) -> None:
	if is_not_blank(detail):
		logger.error(detail)


def raise_400(detail: Any) -> None:
	log_detail(detail)
	raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=detail)


def raise_401(detail: Any, headers: Optional[dict] = None) -> None:
	log_detail(detail)
	if headers is not None:
		raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=detail, headers=headers)
	else:
		raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=detail)


def raise_403() -> None:
	raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail="Unauthorized visit.")


def raise_404(detail: Any = 'Data not found.') -> None:
	log_detail(detail)
	raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=detail)


def raise_406(detail: Any = 'Request not acceptable.') -> None:
	log_detail(detail)
	raise HTTPException(status_code=status.HTTP_406_NOT_ACCEPTABLE, detail=detail)


def raise_500(e: Optional[Exception] = None, detail: Optional[Any] = 'Unpredicted exception occurred.') -> None:
	if e is not None:
		logger.error(e, exc_info=True, stack_info=True)
	log_detail(detail)

	raise HTTPException(
		status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
		detail=detail
	)



from .auth_helper import get_admin_principal, get_any_admin_principal, get_any_principal, get_console_principal, \
	get_principal_by_jwt, get_principal_by_pat, get_super_admin_principal, retrieve_authentication_manager
from .authentication import create_jwt_token, get_admin_principal_by, get_any_admin_principal_by, \
	get_any_principal_by, get_console_principal_by, get_principal_by, get_principal_by_jwt, get_principal_by_pat, \
	get_super_admin_principal_by
from .cors import install_cors

from .exceptions import InitialRestAppException, RestAppException
from .prometheus import install_prometheus
from .rest_app import RestApp
from .settings import RestSettings
